/*
 * cpu-pic30.c -- BFD library support routines for the PIC30 architecture.
 * Copyright (C) 2000 Free Software Foundation, Inc.
 * Contributed by Microchip Corporation.
 * Written by Tracy A. Kuhrt
 *
 * This file is part of BFD, the Binary File Descriptor library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "bfd.h"
#include "sysdep.h"
#include "libbfd.h"
#include "coff-pic30.h"
#include "pic30-utils.h"

/******************************************************************************/

/* prototypes */
const bfd_arch_info_type * pic30_get_machine (const char * const name);
int pic30_proc_family(const bfd_arch_info_type *);
int pic30_is_dsp_machine(const bfd_arch_info_type *);
int pic30_is_eedata_machine(const bfd_arch_info_type *);
int pic30_is_dma_machine(const bfd_arch_info_type *);

const bfd_arch_info_type * global_PROCESSOR = NULL;

#define EOL ,
#define ARCH(machine, printable_name, default_machine, next, flags) \
{ 16,                     \
  24,                     \
  16,                      \
  bfd_arch_pic30,         \
  machine,                \
  "pic30",                \
  printable_name,         \
  1,                      \
  default_machine,        \
  bfd_default_compatible, \
  bfd_default_scan,       \
  next }

static const bfd_arch_info_type arch_info_struct[] =
{
#include "pic30-procs.h"
};

const bfd_arch_info_type bfd_pic30_arch = 
   ARCH ( 0,                 "30Fxxxx", TRUE,  &arch_info_struct[0], 0 ) ;


/*
** Lookup a machine description
*/
const bfd_arch_info_type *
pic30_get_machine (name)
   const char * const name;
{
   const bfd_arch_info_type * p;

   for (p = bfd_pic30_arch.next; p != NULL; p = p->next)
   {
      if (strcasecmp (name, p->printable_name) == 0)
         break;
   } /* for p */

   return p;
}


/*
** Lookup a processor family
*/
int
pic30_proc_family(const bfd_arch_info_type *proc)
{
  int rc = -1;

  if (proc == NULL)
    return rc;

#undef ARCH
#undef EOL
#define ARCH(machine, printable_name, default_machine, next, flags) \
    case machine: \
      rc = (flags) & (FAMILY_MASK); \
      break;

  switch (proc->mach)
  {
 #include "pic30-procs.h"

    default:
      rc = -1;
      break;
  } /* switch */

  return rc;
}


/*
** Query machine for DSP support
*/
int
pic30_is_dsp_machine(const bfd_arch_info_type *proc)
{
  int rc = 1;

  if (proc == NULL)    /* if no processor has been specified,  */
    return rc;         /*  assume it supports DSP instructions */

#undef ARCH
#undef EOL
#define ARCH(machine, printable_name, default_machine, next, flags) \
    case machine: \
      rc = (flags) & F_DSP; \
      break;

  switch (proc->mach)
  {
 #include "pic30-procs.h"

    default:
      rc = 1;
      break;
  } /* switch */

  return rc;
}


/*
** Query machine for EEPROM support
*/
int
pic30_is_eedata_machine(const bfd_arch_info_type *proc)
{
  int rc = 1;

  if (proc == NULL)    /* if no processor has been specified,  */
    return rc;         /*  assume it supports EEDATA */

#undef ARCH
#undef EOL
#define ARCH(machine, printable_name, default_machine, next, flags) \
    case machine: \
      rc = (flags) & F_EEDATA; \
      break;

  switch (proc->mach)
  {
 #include "pic30-procs.h"

    default:
      rc = 1;
      break;
  } /* switch */

  return rc;
}


/*
** Query machine for DMA support
*/
int
pic30_is_dma_machine(const bfd_arch_info_type *proc)
{
  int rc = 1;

  if (proc == NULL)    /* if no processor has been specified,  */
    return rc;         /*  assume it supports DMA */

#undef ARCH
#undef EOL
#define ARCH(machine, printable_name, default_machine, next, flags) \
    case machine: \
      rc = (flags) & F_DMA; \
      break;

  switch (proc->mach)
  {
 #include "pic30-procs.h"

    default:
      rc = 1;
      break;
  } /* switch */

  return rc;
}


int
pic30_display_as_program_memory_p(asection *sec)
{
  int fDisplay;

    fDisplay = ((sec->flags & SEC_CODE));

  return(fDisplay);
}

int
pic30_display_as_data_memory_p(asection *sec)
{
  int fDisplay;

  if ((sec->owner->flags & EXEC_P) == 0) /* unlinked object */
    fDisplay = !PIC30_IS_CODE_ATTR(sec);
  else
    fDisplay = (!PIC30_IS_CODE_ATTR(sec) &&
                !PIC30_IS_PSV_ATTR(sec) &&
                !PIC30_IS_EEDATA_ATTR(sec));

  return(fDisplay);
}

int
pic30_display_as_readonly_memory_p(asection *sec)
{
  int fDisplay;

    fDisplay = ((PIC30_IS_PSV_ATTR(sec) || PIC30_IS_EEDATA_ATTR(sec))
                && ((sec->owner->flags & EXEC_P) == EXEC_P));

  return(fDisplay);
}
